@@ -5,8 +5,8 @@
 ;; Author: Omar Antolín Camarena <omar@matem.unam.mx>, Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Omar Antolín Camarena <omar@matem.unam.mx>, Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2020
-;; Version: 2.1
-;; Package-Requires: ((emacs "28.1") (compat "30"))
+;; Version: 2.4
+;; Package-Requires: ((emacs "29.1") (compat "30"))
 ;; URL: https://github.com/minad/marginalia
 ;; Keywords: docs, help, matching, completion
 
@@ -84,9 +84,6 @@ attribute information.  For Tramp paths, the protocol is
 displayed instead."
   :type '(repeat regexp))
 
-(define-obsolete-variable-alias 'marginalia-annotator-registry
-  'marginalia-annotators "2.0")
-
 (defcustom marginalia-annotators
   (mapcar
    (lambda (x) (append x (list 'builtin 'none)))
@@ -109,6 +106,7 @@ displayed instead."
      (bookmark ,#'marginalia-annotate-bookmark)
      (file ,#'marginalia-annotate-file)
      (project-file ,#'marginalia-annotate-project-file)
+     (project-buffer ,#'marginalia-annotate-buffer)
      (buffer ,#'marginalia-annotate-buffer)
      (library ,#'marginalia-annotate-library)
      (theme ,#'marginalia-annotate-theme)
@@ -163,7 +161,7 @@ matched case-sensitively."
 
 (defcustom marginalia-command-categories
   `((,#'imenu . imenu)
-    (,'recentf-open . file) ;; Available only on Emacs 29.
+    (,#'recentf-open . file)
     (,#'where-is . command))
   "Associate commands with a completion category.
 The value of `this-command' is used as key for the lookup."
@@ -675,10 +673,9 @@ keybinding since CAND includes it."
                   (abbrev-table-p val)))
          (propertize "#<abbrev-table>" 'face 'marginalia-value))
         ((pred char-table-p) (propertize "#<char-table>" 'face 'marginalia-value))
-        ;; Emacs 29 comes with callable objects or object closures (OClosures)
-        ((guard (and (fboundp 'oclosure-type) (oclosure-type val)))
-         (format (propertize "#<oclosure %s>" 'face 'marginalia-function)
-                 (and (fboundp 'oclosure-type) (oclosure-type val))))
+        ;; Callable objects or object closures (OClosures)
+        ((guard (oclosure-type val))
+         (format (propertize "#<oclosure %s>" 'face 'marginalia-function) (oclosure-type val)))
         ((pred byte-code-function-p) (propertize "#<byte-code-function>" 'face 'marginalia-function))
         ((and (pred functionp) (pred symbolp))
          ;; We are not consistent here, values are generally printed
@@ -906,7 +903,10 @@ The string is transformed according to `marginalia--bookmark-type-transforms'."
 
 (defun marginalia-annotate-buffer (cand)
   "Annotate buffer CAND with modification status, file name and major mode."
-  (when-let ((buffer (get-buffer cand)))
+  ;; Emacs 31: `project--read-project-buffer' uses `uniquify-get-unique-names'
+  (when-let ((buffer (or (and (stringp cand)
+                              (get-text-property 0 'uniquify-orig-buffer cand))
+                         (get-buffer cand))))
     (if (buffer-live-p buffer)
         (marginalia--fields
          ((marginalia--buffer-status buffer))
@@ -935,7 +935,8 @@ component of a full file path."
 The return value is a string describing the remote location,
 e.g., the protocol."
   (save-match-data
-    (setq file (substitute-in-file-name file))
+    (setq file (let (file-name-handler-alist)
+                 (substitute-in-file-name file)))
     (cl-loop for r in marginalia-remote-file-regexps
              if (string-match r file)
              return (or (match-string 1 file) "remote"))))
@@ -1129,10 +1130,6 @@ These annotations are skipped for remote paths."
       (put-text-property 0 1 'marginalia--library-doc doc file))
     doc))
 
-(defun marginalia-annotate-theme (cand)
-  "Annotate theme CAND with documentation and path."
-  (marginalia-annotate-library (concat cand "-theme")))
-
 (defun marginalia-annotate-library (cand)
   "Annotate library CAND with documentation and path."
   (setq cand (marginalia--library-name cand))
@@ -1149,6 +1146,15 @@ These annotations are skipped for remote paths."
      ((abbreviate-file-name (file-name-directory file))
       :truncate -0.5 :face 'marginalia-file-name))))
 
+(defun marginalia-annotate-theme (cand)
+  "Annotate theme CAND with documentation and path."
+  (when-let (file (gethash (concat cand "-theme") (marginalia--library-cache)))
+    (marginalia--fields
+     ((marginalia--library-doc file)
+      :truncate 1.0 :face 'marginalia-documentation)
+     ((abbreviate-file-name (file-name-directory file))
+      :truncate -1.0 :face 'marginalia-file-name))))
+
 (defun marginalia-annotate-tab (cand)
   "Annotate named tab CAND with tab index, window and buffer information."
   (when-let ((tabs (funcall tab-bar-tabs-function))
@@ -1350,10 +1356,9 @@ Remember `this-command' for `marginalia-classify-by-command-name'."
   "Get completion metadata."
   (let* ((end (minibuffer-prompt-end))
          (pt (max 0 (- (point) end))))
-    (and (minibufferp)
-         (completion-metadata (buffer-substring-no-properties end (+ end pt))
-                              minibuffer-completion-table
-                              minibuffer-completion-predicate))))
+    (completion-metadata (buffer-substring-no-properties end (+ end pt))
+                         minibuffer-completion-table
+                         minibuffer-completion-predicate)))
 
 (defun marginalia--builtin-annotator-p (md)
   "Builtin annotator available in metadata MD?"
@@ -1393,8 +1398,10 @@ Remember `this-command' for `marginalia-classify-by-command-name'."
                      if (or (not (eq fun 'builtin)) (marginalia--builtin-annotator-p md))
                      collect
                      (vector
-                      (capitalize (replace-regexp-in-string
-                                   ".*?-+annotate-+" "" (symbol-name fun)))
+                      (thread-last (symbol-name fun)
+                                   (replace-regexp-in-string ".*?-+annotate-+" "")
+                                   (replace-regexp-in-string "-+" " ")
+                                   capitalize)
                       (let ((i i))
                         (lambda ()
                           (interactive)
